// -*-c++-*-

/*!
  \file pass.h
  \brief pass object Header File.
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 3 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef PASS_H
#define PASS_H

#include "cooperative_action.h"

#include <rcsc/geom/vector_2d.h>
#include <rcsc/game_time.h>
#include <rcsc/types.h>

#include <boost/shared_ptr.hpp>

class Pass
    : public CooperativeAction {
private:

    double M_ball_speed; //!< first ball speed

    // double M_receive_point_error;

public:

    Pass( const rcsc::GameTime & start_time,
          const int passer,
          const rcsc::Vector2D & start_point );

    Pass( const rcsc::GameTime & start_time,
          const int passer,
          const int receiver,
          const rcsc::Vector2D & start_point,
          const rcsc::Vector2D & receive_point,
          const double & ball_speed );

    virtual
    const char * name() const;

    const double & ballSpeed() const
      {
          return M_ball_speed;
      }

    void setReceiver( const int receiver,
                      const rcsc::Vector2D & receive_point,
                      const double & ball_speed );

};

#endif
